/*
  File name: CCytadelaMain.cpp
  Copyright: (C) 2005 - 2007 Tomasz Kazmierczak

  Creation date: 15.07.2005 20:24
  Last modification date: 28.08.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#include "CCytadelaMain.h"
#include "CCytadelaMenu.h"
#include <memory.h>
#include <dirent.h>
#include <sys/types.h>
#include <iostream>
#include <stdio.h>

using namespace std;

CCytadelaMain::CCytadelaMain(const char *szSysDirPath)
{
	memset(mState.wAmmo, 0, 6 * sizeof(mState.wAmmo));
	mLevel = NO_LEVEL;
	mState.bDifficult = true;
	mState.dwTotalTime = 0;
	mState.wEnergy = 0;
	mState.wKilled = 0;
	mState.btBomb = 0;
	mSDL = CSDLClass::instance();
	mLocInterface = ILocalization::instance();
	strcpy(mSysDirPath, szSysDirPath);

	mLevels[0] = "obj/Podziemia1.3dg";
	mLevels[1] = "obj/Podziemia2.3dg";
	mLevels[2] = "obj/Elektrownia1.3dg";
	mLevels[3] = "obj/Elektrownia2.3dg";
	mLevels[4] = "obj/Elektrownia3.3dg";
	mLevels[5] = "obj/Magazyny1.3dg";
	mLevels[6] = "obj/Magazyny2.3dg";
	mLevels[7] = "obj/Magazyny3.3dg";
	mLevels[8] = "obj/Hangar1.3dg";
	mLevels[9] = "obj/Hangar2.3dg";
	mLevels[10] = "obj/Hangar3.3dg";
	mLevels[11] = "obj/Laboratoria1.3dg";
	mLevels[12] = "obj/Laboratoria2.3dg";
	mLevels[13] = "obj/Laboratoria3.3dg";
	mLevels[14] = "obj/Kanaly1.3dg";
	mLevels[15] = "obj/Kanaly2.3dg";
	mLevels[16] = "obj/Kanaly3.3dg";
	mLevels[17] = "obj/Wiezienie1.3dg";
	mLevels[18] = "obj/Wiezienie2.3dg";
	mLevels[19] = "obj/Wiezienie3.3dg";
	mLevels[20] = "obj/Centrum.3dg";
	mLevels[21] = "obj/PodziemiaT.3dg";
	mLevels[22] = "obj/MagazynyT1.3dg";
	mLevels[23] = "obj/MagazynyT2.3dg";
	mLevels[24] = "obj/LaboratoriaT.3dg";
	mLevels[25] = "obj/KanalyT.3dg";

	mCplxStatus[SEWERS] = ACTIVE;
	mCplxStatus[LABORATORY] = ACTIVE;
	mCplxStatus[POWER_PLANT] = ACTIVE;
	mCplxStatus[STORES] = INACTIVE;
	mCplxStatus[HANGAR] = INACTIVE;
	mCplxStatus[PRISON] = INACTIVE;
	mCplxStatus[CENTER] = INACTIVE;

	mBeginLevel[SEWERS] = LEVEL06A;
	mBeginLevel[LABORATORY] = LEVEL05A;
	mBeginLevel[POWER_PLANT] = LEVEL02A;
	mBeginLevel[STORES] = LEVEL03A;
	mBeginLevel[HANGAR] = LEVEL04A;
	mBeginLevel[PRISON] = LEVEL07A;
	mBeginLevel[CENTER] = LEVEL08A;
	mCplxCounter = 1;
	mActualComplex = 0;
	mState.bExitGame = false;
}

void CCytadelaMain::reset(CCytadelaConfig *pConfig)
{
	mCplxStatus[SEWERS] = ACTIVE;
	mCplxStatus[LABORATORY] = ACTIVE;
	mCplxStatus[POWER_PLANT] = ACTIVE;
	mCplxStatus[STORES] = INACTIVE;
	mCplxStatus[HANGAR] = INACTIVE;
	mCplxStatus[PRISON] = INACTIVE;
	mCplxStatus[CENTER] = INACTIVE;
	memset(mState.wAmmo, 0, 6 * sizeof(uint16_t));
	mCplxCounter = 1;
	mActualComplex = 0;
	mState.btBomb = 0;
	mState.wEnergy = 150;
	mState.wWeapons = 0;
	mState.wKilled = 0;
	mState.dwTotalTime = 0;
	mState.bExitGame = false;
	mState.bDifficult = pConfig->isDifficult();
	mState.bVisionNoise = true;
	for(int32_t i = 0; i < TEXT_CLASSES_COUNT; ++i)
		mTexts[i].resetTextCounter();
}

void CCytadelaMain::setLevelTexts()
{
	mSaveNames[0] = mLocInterface->getText(LOC_SAVE_DUNGEONS_1);
	mSaveNames[1] = mLocInterface->getText(LOC_SAVE_DUNGEONS_2);
	mSaveNames[2] = mLocInterface->getText(LOC_SAVE_POWER_PLANT_1);
	mSaveNames[3] = mLocInterface->getText(LOC_SAVE_POWER_PLANT_2);
	mSaveNames[4] = mLocInterface->getText(LOC_SAVE_POWER_PLANT_3);
	mSaveNames[5] = mLocInterface->getText(LOC_SAVE_STORES_1);
	mSaveNames[6] = mLocInterface->getText(LOC_SAVE_STORES_2);
	mSaveNames[7] = mLocInterface->getText(LOC_SAVE_STORES_3);
	mSaveNames[8] = mLocInterface->getText(LOC_SAVE_HANGAR_1);
	mSaveNames[9] = mLocInterface->getText(LOC_SAVE_HANGAR_2);
	mSaveNames[10] = mLocInterface->getText(LOC_SAVE_HANGAR_3);
	mSaveNames[11] = mLocInterface->getText(LOC_SAVE_LABORATORY_1);
	mSaveNames[12] = mLocInterface->getText(LOC_SAVE_LABORATORY_2);
	mSaveNames[13] = mLocInterface->getText(LOC_SAVE_LABORATORY_3);
	mSaveNames[14] = mLocInterface->getText(LOC_SAVE_SEWERS_1);
	mSaveNames[15] = mLocInterface->getText(LOC_SAVE_SEWERS_2);
	mSaveNames[16] = mLocInterface->getText(LOC_SAVE_SEWERS_3);
	mSaveNames[17] = mLocInterface->getText(LOC_SAVE_PRISON_1);
	mSaveNames[18] = mLocInterface->getText(LOC_SAVE_PRISON_2);
	mSaveNames[19] = mLocInterface->getText(LOC_SAVE_PRISON_3);
	mSaveNames[20] = mLocInterface->getText(LOC_SAVE_CENTER_1);

	mCplxNames[SEWERS] = mLocInterface->getText(LOC_CPLX_SEWERS);
	mCplxNames[LABORATORY] = mLocInterface->getText(LOC_CPLX_LABORATORY);
	mCplxNames[POWER_PLANT] = mLocInterface->getText(LOC_CPLX_POWER_PLANT);
	mCplxNames[STORES] = mLocInterface->getText(LOC_CPLX_STORES);
	mCplxNames[HANGAR] = mLocInterface->getText(LOC_CPLX_HANGAR);
	mCplxNames[PRISON] = mLocInterface->getText(LOC_CPLX_PRISON);
	mCplxNames[CENTER] = mLocInterface->getText(LOC_CPLX_CENTER);

	mCplxStatusName[ACTIVE] = mLocInterface->getText(LOC_CPLX_STATUS_UNEXPLORED);
	mCplxStatusName[INACTIVE] = mLocInterface->getText(LOC_CPLX_STATUS_NO_ACCESS);
	mCplxStatusName[FINISHED] = mLocInterface->getText(LOC_CPLX_STATUS_EXPLORED);
	mCplxStatusName[CENTERINACTIVE] = mLocInterface->getText(LOC_CPLX_STATUS_NO_ACCESS);
}

bool CCytadelaMain::configScreen(CCytadelaConfig *pConfig)
{
	fprintf(stdout, "------------------------------------------------------------\n");
	fprintf(stdout, "Reading the OpenGL vrsion and extensions...   \n");
	//switch to OpenGL mode to check the version
	if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_3D))
	{
		fprintf(stdout, "\nError: Could not initialize OpenGL!\nExiting...\n");
		return false;
	}

	std::string version((const char *)glGetString(GL_VERSION));
	std::string renderer((const char *)glGetString(GL_RENDERER));
	std::string extensions((const char *)glGetString(GL_EXTENSIONS));

	fprintf(stdout, "Done\n");
	fprintf(stdout, "Switching to 2D video mode...   ");
	//now switch to the 2D mode
	if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_2D))
	{
		fprintf(stdout, "\nError: Could not initialize SDL Video mode!\nExiting...\n");
		return false;
	}
	fprintf(stdout, "Ok!\n");

	fprintf(stdout, "Initializing configuration menu...\n");
	uint32_t uiImgNum, uiFontImgNum;
	//load startup screen
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "config.tga", &uiImgNum))
	{
		fprintf(stdout, "\nError: Could not load menu image!\nExiting...\n");
		return false;
	}
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "configfont.tga", &uiFontImgNum))
	{
		mSDL->mImgArray.freeIA();
		fprintf(stdout, "\nError: Could not load font image!\nExiting...\n");
		return false;
	}
	int32_t iSound = mSDL->loadSound("sounds/configmenu.wav");
	if(CSDL_LOAD_SOUND_FAILED == iSound)
	{
		mSDL->mImgArray.freeIA();
		fprintf(stdout, "\nError: Could not load menu sound!\nExiting...\n");
		return false;
	}
	fprintf(stdout, "Done\n");

	mSDL->playSoundFadeIn(iSound, -1, 500);
	mSDL->displayImage(uiImgNum, 0, 0, 0);
	mSDL->displayText(uiFontImgNum, "  CYTADELA CONFIGURATION", 112, 90);

	fprintf(stdout, "Cheching the OpenGL version and extensions...\n\n");
	//OpenGL version and extensions
	char szVerNum[4];
	memcpy(szVerNum, version.c_str(), 3 * sizeof(char));
	szVerNum[3] = '\0';
	float fOGLVer = atof(szVerNum);

	mSDL->displayText(uiFontImgNum, "GL VERSION: ", 80, 130);
	mSDL->displayText(uiFontImgNum, version.c_str(), 272, 130);
	mSDL->displayText(uiFontImgNum, "GL RENDERER: ", 64, 150);
	mSDL->displayText(uiFontImgNum, renderer.c_str(), 272, 150);
	//check for version and extensions
	bool bOK = true;
	if(fOGLVer < 1.4f)
	{
		if(fOGLVer < 1.2f)
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "CYTADELA REQUIRES AT LEAST OPENGL 1.2!", 16, 230);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 250);
			goto AfterCheck;
		}
		mSDL->displayText(uiFontImgNum, "NEEDED OPENGL EXTENSIONS: ", 96, 190);
		mSDL->displayText(uiFontImgNum, "GL_ARB_TEXTURE_MIRRORED_REPEAT: ", 8, 210);
		if(checkGLExtension("GL_ARB_texture_mirrored_repeat", (const GLubyte *)extensions.c_str()))
			mSDL->displayText(uiFontImgNum, "YES", 536, 210);
		else
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "NO", 536, 210);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 320);
		}
		if(fOGLVer == 1.3f)
			goto AfterCheck;
		mSDL->displayText(uiFontImgNum, "GL_ARB_TEXTURE_BORDER_CLAMP: ", 8, 230);
		if(checkGLExtension("GL_ARB_texture_border_clamp", (const GLubyte *)extensions.c_str()))
			mSDL->displayText(uiFontImgNum, "YES", 536, 230);
		else
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "NO", 536, 230);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 320);
		}
		mSDL->displayText(uiFontImgNum, "GL_ARB_MULTITEXTURE: ", 8, 250);
		if(checkGLExtension("GL_ARB_multitexture", (const GLubyte *)extensions.c_str()))
			mSDL->displayText(uiFontImgNum, "YES", 536, 250);
		else
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "NO", 536, 250);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 320);
		}
		mSDL->displayText(uiFontImgNum, "GL_ARB_TEXTURE_ENV_ADD: ", 8, 270);
		if(checkGLExtension("GL_ARB_texture_env_add", (const GLubyte *)extensions.c_str()))
			mSDL->displayText(uiFontImgNum, "YES", 536, 270);
		else
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "NO", 536, 270);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 320);
		}
		mSDL->displayText(uiFontImgNum, "GL_ARB_TEXTURE_ENV_COMBINE: ", 8, 290);
		if(checkGLExtension("GL_ARB_texture_env_combine", (const GLubyte *)extensions.c_str()))
			mSDL->displayText(uiFontImgNum, "YES", 536, 290);
		else
		{
			bOK = false;
			mSDL->displayText(uiFontImgNum, "NO", 536, 290);
			mSDL->displayText(uiFontImgNum, "PRESS \'ENTER\' OR \'ESC\' TO EXIT", 80, 320);
		}
	}

	fprintf(stdout, "OpenGL version: %s\nOpenGL renderer: %s\n", version.c_str(), renderer.c_str());
	AfterCheck:

	if(not bOK)
	{
		mSDL->drawFrame();
		SDLKey eKey;
		while(true)
		{
			eKey = mSDL->readInputMenu();
			if(eKey == SDLK_RETURN or eKey == SDLK_ESCAPE)
				break;
		}
		fprintf(stdout, "\nOpenGL version to low or missing some needed GL extensions - exiting...\n");
		mSDL->stopSound(iSound, 500);
		mSDL->delay(500);
		mSDL->freeSounds();
		mSDL->mImgArray.freeIA();
		return false;
	}

	//localizations
	fprintf(stdout, "Checking for available localizations (at \"locale/\")...   ");

	DIR *pDir = opendir("locale");
	if(not pDir)
	{
		fprintf(stdout, "\nError: could not open \"locale/\" directory!\nExiting...\n");
		mSDL->freeSounds();
		mSDL->mImgArray.freeIA();
		return false;
	}

	//+1 because the first will be the CYTADELA_DEFAULT_LOCALIZATION
	long lLocalesCount = 1;
	long lFilesCount = 0;;
	while(true)
	{
		dirent *pDirEntry = readdir(pDir);
		if(not pDirEntry)
			break;
		lFilesCount++;
		if(strstr(pDirEntry->d_name, ".loc"))
			lLocalesCount++;
	}
	rewinddir(pDir);
	//strings to store filenames
	string *szEntryNames = new string[lLocalesCount];
	if(not szEntryNames)
	{
		fprintf(stdout, "\nError: out of memory!\nExiting...\n");
		mSDL->freeSounds();
		mSDL->mImgArray.freeIA();
		closedir(pDir);
		return false;
	}
	szEntryNames[0] = CYTADELA_DEFAULT_LOCALIZATION;

	for(long l = 0, m = 1; l < lFilesCount; ++l)
	{
		dirent *pDirEntry = readdir(pDir);
		if(strstr(pDirEntry->d_name, ".loc"))
		{
			szEntryNames[m] = pDirEntry->d_name;
			//remove the extension from the filename
			string::size_type stPos = szEntryNames[m].find(".loc");
			szEntryNames[m].replace(stPos, 1, 1, '\0');
			m++;
		}
	}

	closedir(pDir);
	fprintf(stdout, "Done\n");

	//find the position of localization read from config file (set it initially
	//to the default value in case it is not found in the config)
	long lChosenLocalization = 0;
	for(long l = 0; l < lLocalesCount; ++l)
	{
		if(szEntryNames[l].find(pConfig->getLocalization(), 0) != string::npos)
		{
			lChosenLocalization = l;
			break;
		}
	}

	fprintf(stdout, "Entering the config menu main loop...\n");
	SDLKey eKey;
	char szLocalization[FILENAME_MAX];
	mSDL->clearEventQueue();
	while(true)
	{
		eKey = mSDL->readInputMenu();
		//if user chose to exit
		if(eKey == SDLK_ESCAPE)
		{
			delete[] szEntryNames;
			fprintf(stdout, "Program terminated by user - exiting...\n");
			mSDL->stopSound(iSound, 500);
			mSDL->delay(500);
			mSDL->freeSounds();
			mSDL->mImgArray.freeIA();
			return false;
		}
		if(eKey == SDLK_RETURN)
		{
			sprintf(szLocalization, szEntryNames[lChosenLocalization].c_str());
			break;
		}
		if(eKey == SDLK_UP)
		{
			lChosenLocalization--;
			if(lChosenLocalization < 0)
				lChosenLocalization = lLocalesCount-1;
		}
		if(eKey == SDLK_DOWN)
		{
			lChosenLocalization++;
			if(lChosenLocalization == lLocalesCount)
				lChosenLocalization = 0;
		}
		//redraw the localization menu
		SDL_Rect sSrcRect = {0, 310, 640, 480};
		mSDL->displayImage(uiImgNum, &sSrcRect, 0, 310);
		mSDL->displayText(uiFontImgNum, "CHOOSE LANGUAGE: ", 112, 370);
		mSDL->displayText(uiFontImgNum, "(CURSOR UP/DOWN)", 112, 390);
		for(long l = 0; l < lLocalesCount; ++l)
		{
			//display only 3 before and 3 after the chosen
			if(l > lChosenLocalization - 4 and l < lChosenLocalization + 4)
			{
				mSDL->displayText(uiFontImgNum, (char *)szEntryNames[l].c_str(), 384,
									370 + (l - lChosenLocalization)*20);
			}
		}
		mSDL->drawFrame();
	}
	delete[] szEntryNames;

	//NOTE: localization can be set only once and it should be done at the begining!
	if(not mLocInterface->loadLocalization(szLocalization, mTexts))
	{
		fprintf(stderr, "Error while loading localization from locale/%s.txt: %s\n"
						"Using default (%s) localization\n", szLocalization,
						getLocErrorDesc(mLocInterface->getLastError()),
						CYTADELA_DEFAULT_LOCALIZATION);
		mSDL->displayImage(uiImgNum, 0, 0, 0);

		mSDL->displayText(uiFontImgNum, "ERROR:", 272, 170);
		mSDL->displayText(uiFontImgNum, "COULD NOT LOAD CHOSEN LOCALIZATION:", 48, 210);
		mSDL->displayText(uiFontImgNum, szLocalization, 272, 230);
		char szErrorText[FILENAME_MAX];
		sprintf(szErrorText, "USING DEFAULT (%s) LOCALIZATION", CYTADELA_DEFAULT_LOCALIZATION);
		mSDL->displayText(uiFontImgNum, szErrorText, 48, 270);
		mSDL->drawFrame();
		mLocInterface->setDefaultLocalization();
		while(SDLK_RETURN != mSDL->readInputMenu());
	}
	else fprintf(stdout, "Chosen localization: %s\n", szLocalization);
	pConfig->setLocalization(szLocalization);

	fprintf(stdout, "Initial check and configuration finished successfuly!\n");
	fprintf(stdout, "------------------------------------------------------------\n");
	mSDL->stopSound(iSound, 500);
	mSDL->delay(500);
	mSDL->freeSounds();
	mSDL->mImgArray.freeIA();
	return true;
}

bool CCytadelaMain::startupScreen()
{
	fprintf(stdout, "Displaying startup screen:\n");
	fprintf(stdout, "Initializing...   \n");
	uint32_t dwImgNum;
	//load startup screen
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "cytadela.tga", &dwImgNum))
		return false;
	if(not mSDL->loadMusic("music/Intro.mod"))
		return false;
	fprintf(stdout, "   Done!\n");
	mSDL->playMusic(1);
	//display it (with animation)
	fprintf(stdout, "Running the animation...\n");
	uint32_t dwY = 0;
	uint32_t dwT0 = mSDL->getTicks();
	uint32_t dwBeginTime = dwT0;
	while(dwY <= 480)
	{
		dwY = (mSDL->getTicks() - dwT0);
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, 480 - dwY);
		mSDL->drawFrame();
	}
	dwY = 0;
	dwT0 = mSDL->getTicks();
	while(dwY <= 50)
	{
		dwY = (mSDL->getTicks() - dwT0) / 10;
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, dwY);
		mSDL->drawFrame();
	}
	dwT0 = mSDL->getTicks();
	while(dwY <= 480)
	{
		dwY = 430 + (mSDL->getTicks() - dwT0) / 6;
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, 480 - dwY);
		mSDL->drawFrame();
	}
	dwY = 0;
	dwT0 = mSDL->getTicks();
	while(dwY <= 10)
	{
		dwY = (mSDL->getTicks() - dwT0) / 30;
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, dwY);
		mSDL->drawFrame();
	}
	dwT0 = mSDL->getTicks();
	while(dwY <= 480)
	{
		dwY = 470 + (mSDL->getTicks() - dwT0) / 20;
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, 480 - dwY);
		mSDL->drawFrame();
	}
	mSDL->displayImage(dwImgNum, 0, 0, 0);
	mSDL->drawFrame();
	fprintf(stdout, "Animation finished.\n");
	mSDL->clearEventQueue();
	//we wait till user presses some key or time elapses
	while(SDLK_RETURN != mSDL->readInputMenu() and (mSDL->getTicks() - dwBeginTime) < 17500);
	fprintf(stdout, "Scrolling down...\n");
	//scroll down
	dwY = 0;
	dwT0 = mSDL->getTicks();
	while(dwY <= 480)
	{
		dwY = (mSDL->getTicks() - dwT0);
		mSDL->blackScreen();
		mSDL->displayImage(dwImgNum, 0, 0, dwY);
		mSDL->drawFrame();
	}
	mSDL->stopMusic();
	mSDL->freeMusic();
	mSDL->mImgArray.freeIA();
	fprintf(stdout, "Done with the startup screen!\n");
	fprintf(stdout, "------------------------------------------------------------\n");
	return true;
}

bool CCytadelaMain::endLevel()
{
	uint32_t dwImgNum, dwFont, dwEndCur, dwSaveCur;
	//load end-level screen, etc.
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "info.tga", &dwImgNum))
		return false;
	//load font for chosen localization
	char szFontPath[FILENAME_MAX];
	sprintf(szFontPath, "tex/fonts/%s.far", mLocInterface->getCurrentLocalization());
	if(not mSDL->mImgArray.loadImg(szFontPath, "endfont.tga", &dwFont))
		return false;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "endcur.tga", &dwEndCur))
		return false;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "savecur.tga", &dwSaveCur))
		return false;
	if(not mSDL->loadMusic("music/EndLevel.mod"))
		return false;
	mSDL->displayImage(dwImgNum, 0, 0, 0);
	mSDL->drawFrame();
	mSDL->playMusic(-1);
	//display text
	char szText[40];
	sprintf(szText, mSaveNames[mLevel]);
	char cLevelNum = szText[strlen(szText) - 1];
	szText[strlen(szText) - 1] = 0;
	sprintf(szText, "%s%s %c %s!", szText, mLocInterface->getText(LOC_END_LEVEL),
			cLevelNum, mLocInterface->getText(LOC_END_COMPLETED));
	mSDL->displayText(dwFont, szText, 626 - 16 * strlen(szText), 92);
	mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_KILLED), 20, 212);
	mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_TIME), 20, 242);
	mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_BOMB), 20, 272);
	sprintf(szText, "%d", mState.wKilled);
	mSDL->displayText(dwFont, szText, 484, 212);
	mState.dwTotalTime /= 1000;
	uint32_t dwHours = mState.dwTotalTime / 3600;
	uint32_t dwMinutes = (mState.dwTotalTime - dwHours * 3600) / 60;
	uint32_t dwSeconds = mState.dwTotalTime - dwHours * 3600 - dwMinutes * 60;
	sprintf(szText, "%d:", dwHours);
	if(dwMinutes >= 10)
		sprintf(szText + 2, "%d:", dwMinutes);
	else
		sprintf(szText + 2, "0%d:", dwMinutes);
	if(dwSeconds >= 10)
		sprintf(szText + 5, "%d", dwSeconds);
	else
		sprintf(szText + 5, "0%d", dwSeconds);
	mSDL->displayText(dwFont, szText, 484, 242);
	sprintf(szText, "%d", mState.btBomb);
	mSDL->displayText(dwFont, szText, 484, 272);
	mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_SAVE_QUESTION), 52, 362);
	mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_NO_YES), 212, 398);
	mSDL->drawFrame();
	bool bSave = false;
	uint32_t pImgNums[] = {dwImgNum, dwFont, dwEndCur, dwSaveCur};
	mSDL->clearEventQueue();
	while(true)
	{
		SDLKey eKey = mSDL->readInputMenu();
		if(eKey == SDLK_ESCAPE)
		{
			mState.bExitGame = true;
			break;
		}
		if(eKey == SDLK_RIGHT or eKey == SDLK_LEFT)
			bSave = not bSave;
		if(eKey == SDLK_RETURN or eKey == SDLK_SPACE)
		{
			if(bSave)
			{
				if(not saveScreen(pImgNums))
				{
					mSDL->stopMusic();
					mSDL->freeMusic();
					mSDL->mImgArray.freeIA();
					return false;
				}
				break;
			}
			else break;
		}
		//animate cursor
		SDL_Rect sSrcRect = {0, 390, 640, 90};
		mSDL->displayImage(dwImgNum, &sSrcRect, 0, 390);
		mSDL->displayText(dwFont, mLocInterface->getText(LOC_END_NO_YES), 212, 398);
		sSrcRect.x = 0;
		sSrcRect.y = ((mSDL->getTicks() / 500) & 1) * 34;
		sSrcRect.w = 52;
		sSrcRect.h = 34;
		mSDL->displayImage(dwEndCur, &sSrcRect, 210 + bSave * 192, 394);
		mSDL->drawFrame();
	}
	mSDL->stopMusic();
	mSDL->blackScreen();
	mSDL->drawFrame();
	mSDL->freeMusic();
	mSDL->mImgArray.freeIA();
	return true;
}

bool CCytadelaMain::saveScreen(uint32_t *pImgNums)
{
	bool bEndLoop = false;
	uint32_t dwNum = 0;
	CCytadelaSave clSave;
	if(not clSave.readSaveFile(mSysDirPath))
		return false;
	mSDL->clearEventQueue();
	while(not bEndLoop)
	{
		mSDL->displayImage(pImgNums[0], 0, 0, 0);
		mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_END_SAVE_CHOOSE_POS),
							132, 72);
		for(int32_t i = 0; i < 5; ++i)
		{
			Levels eLevel = clSave.mSave[i].eLevel;
			if(eLevel == NO_LEVEL)
				mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_SAVE_EMPTY),
									196, 132 + 40 * i);
			else
				mSDL->displayText(pImgNums[1], mSaveNames[eLevel], 196, 132 + 40 * i);
		}
		mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_END_SAVE_QUIT), 244, 352);
		//animate cursor
		SDL_Rect sSrcRect = {0, ((mSDL->getTicks() / 500) & 1) * 34, 260, 34};
		if(dwNum != 5)
			mSDL->displayImage(pImgNums[3], &sSrcRect, 192, 128 + dwNum * 40);
		else
			mSDL->displayImage(pImgNums[3], &sSrcRect, 192, 348);
		mSDL->drawFrame();
		//read input
		SDLKey eKey = mSDL->readInputMenu();
		switch(eKey)
		{
		case SDLK_UP:
			{
				if(dwNum == 0) dwNum = 5;
				else dwNum--;
			}
			break;
		case SDLK_DOWN:
			{

				if(dwNum == 5) dwNum = 0;
				else dwNum++;
			}
			break;
		case SDLK_ESCAPE:
		case SDLK_RETURN:
		case SDLK_SPACE:
			{
				if(dwNum < 5)
				{
					//save all needed info
					for(int32_t i = 0; i < 7; ++i)
						clSave.mSave[dwNum].eCplxStatus[i] = mCplxStatus[i];
					for(int32_t i = 0; i < 6; ++i)
						clSave.mSave[dwNum].wAmmo[i] = mState.wAmmo[i];
					clSave.mSave[dwNum].eLevel = mLevel;
					clSave.mSave[dwNum].bDifficult = mState.bDifficult;
					clSave.mSave[dwNum].btActualComplex = mActualComplex;
					clSave.mSave[dwNum].btBomb = mState.btBomb;
					clSave.mSave[dwNum].wCplxCounter = mCplxCounter;
					clSave.mSave[dwNum].wEnergy = mState.wEnergy;
					clSave.mSave[dwNum].wWeapons = mState.wWeapons;
					clSave.mSave[dwNum].bVisionNoise = mState.bVisionNoise;
					memcpy(clSave.mSave[dwNum].wAmmo, mState.wAmmo, 6 * sizeof(uint16_t));
					if(not clSave.writeSaveFile(mSysDirPath))
						return false;
					//display new save name in list
					mSDL->displayImage(pImgNums[0], 0, 0, 0);
					mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_END_SAVE_CHOOSE_POS), 132, 72);
					for(int32_t i = 0; i < 5; ++i)
					{
						Levels eLevel = clSave.mSave[i].eLevel;
						if(eLevel == NO_LEVEL)
							mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_SAVE_EMPTY),
																					196, 132 + 40 * i);
						else
							mSDL->displayText(pImgNums[1], mSaveNames[eLevel], 196, 132 + 40 * i);
					}
					mSDL->displayText(pImgNums[1], mLocInterface->getText(LOC_END_SAVE_QUIT), 244, 352);
					mSDL->drawFrame();
					mSDL->delay(1000);
					//display "ok!"
					mSDL->displayImage(pImgNums[0], 0, 0, 0);
					mSDL->displayText(pImgNums[1], "OK!", 308, 192);
					mSDL->drawFrame();
					mSDL->delay(1500);
				}
				bEndLoop = true;
			}
			break;
		default:
			break;
		}
	}
	return true;
}

bool CCytadelaMain::chooseComplex()
{
	uint32_t dwMapImg, dwCursorImg, dwFontImg;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "map.tga", &dwMapImg))
		return false;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "mapcur.tga", &dwCursorImg))
		return false;
	//load font for chosen localization
	char szFontPath[FILENAME_MAX];
	sprintf(szFontPath, "tex/fonts/%s.far", mLocInterface->getCurrentLocalization());
	if(not mSDL->mImgArray.loadImg(szFontPath, "mapfont.tga", &dwFontImg))
		return false;
	if(not mSDL->loadMusic("music/Map.mod"))
		return false;
	mSDL->displayImage(dwMapImg, 0, 0, 0);
	mSDL->drawFrame();
	mSDL->playMusic(-1);
	uint32_t dwNum = mActualComplex;
	uint32_t dwLastTime = mSDL->getTicks();
	uint32_t dwCursorFrame = 0;
	bool bEndLoop = false;
	char szText[56];
	mSDL->clearEventQueue();
	while(not bEndLoop)
	{
		mSDL->displayImage(dwMapImg, 0, 0, 0);
		//animate cursor
		animateMapCursor(dwCursorImg, &dwLastTime, &dwCursorFrame, dwNum);
		sprintf(szText, "%s:%s", mLocInterface->getText(LOC_CPLX_MENU_COMPLEX),
								mCplxNames[dwNum]);
		mSDL->displayText(dwFontImg, szText, 8, 426);
		sprintf(szText, "%s:%s", mLocInterface->getText(LOC_CPLX_MENU_STATUS),
								mCplxStatusName[mCplxStatus[dwNum]]);
		mSDL->displayText(dwFontImg, szText, 344, 426);
		mSDL->drawFrame();
		//read input
		SDLKey eKey = mSDL->readInputMenu();
		switch(eKey)
		{
		case SDLK_DOWN:
		case SDLK_LEFT:
			{
				if(dwNum == 0) dwNum = 6;
				else dwNum--;
			}
			break;
		case SDLK_UP:
		case SDLK_RIGHT:
			{
				if(dwNum == 6) dwNum = 0;
				else dwNum++;
			}
			break;
		case SDLK_ESCAPE:
			{
				mState.bExitGame = true;
				bEndLoop = true;
			}
			break;
		case SDLK_RETURN:
		case SDLK_SPACE:
			{
				switch(mCplxStatus[dwNum])
				{
				case ACTIVE:
					{
						mLevel = mBeginLevel[dwNum];
						mCplxStatus[dwNum] = FINISHED;
						mActualComplex = dwNum;
						mCplxCounter++;
						switch(mCplxCounter)
						{
						case 2:
							mCplxStatus[STORES] = ACTIVE;
							break;
						case 3:
							mCplxStatus[HANGAR] = ACTIVE;
							break;
						case 4:
							mCplxStatus[PRISON] = ACTIVE;
							break;
						case 5:
							mCplxStatus[CENTER] = CENTERINACTIVE;
							break;
						case 7:
							{
								//if don't have all bomb parts than player can't
								//go into the Center
								if(mState.btBomb == 6)
									mCplxStatus[CENTER] = ACTIVE;
							}
							break;
						}
						uint32_t dwT0 = mSDL->getTicks();
						while(mSDL->getTicks() - dwT0 < 2000)
						{
							mSDL->displayImage(dwMapImg, 0, 0, 0);
							animateMapCursor(dwCursorImg, &dwLastTime,
									&dwCursorFrame, dwNum);
							sprintf(szText, "%s %s", mLocInterface->getText(LOC_CPLX_MENU_CHOSEN),
																				mCplxNames[dwNum]);
							//if power station - change 'A' to 'e' (for Polish localization;)
							if(dwNum == 4 and strcmp(mLocInterface->getCurrentLocalization(),
												CYTADELA_DEFAULT_LOCALIZATION) == 0)
								szText[19] = 'e';
							mSDL->displayText(dwFontImg, szText, 184, 426);
							mSDL->drawFrame();
						}
						bEndLoop = true;
					}
					break;
				case INACTIVE:
					{
						uint32_t dwT0 = mSDL->getTicks();
						while(mSDL->getTicks() - dwT0 < 2000)
						{
							mSDL->displayImage(dwMapImg, 0, 0, 0);
							animateMapCursor(dwCursorImg, &dwLastTime, &dwCursorFrame, dwNum);
							mSDL->displayText(dwFontImg, mLocInterface->getText(LOC_CPLX_MENU_NO_ACCESS_YET),
																										8, 426);
							mSDL->drawFrame();
						}
						mSDL->displayText(dwFontImg, szText, 8, 426);
						mSDL->clearEventQueue();
					}
					break;
				case FINISHED:
					{
						uint32_t dwT0 = mSDL->getTicks();
						while(mSDL->getTicks() - dwT0 < 2000)
						{
							mSDL->displayImage(dwMapImg, 0, 0, 0);
							animateMapCursor(dwCursorImg, &dwLastTime, &dwCursorFrame, dwNum);
							mSDL->displayText(dwFontImg, mLocInterface->getText(LOC_CPLX_MENU_ALREADY_EXPLORED),
																											8, 426);
							mSDL->drawFrame();
						}
						mSDL->displayText(dwFontImg, szText, 8, 426);
						mSDL->clearEventQueue();
					}
					break;
				case CENTERINACTIVE:
					{
						uint32_t dwT0 = mSDL->getTicks();
						while(mSDL->getTicks() - dwT0 < 2000)
						{
							mSDL->displayImage(dwMapImg, 0, 0, 0);
							animateMapCursor(dwCursorImg, &dwLastTime, &dwCursorFrame, dwNum);
							mSDL->displayText(dwFontImg, mLocInterface->getText(LOC_CPLX_MENU_NEED_WHOLE_BOMB),
																											8, 426);
							mSDL->drawFrame();
						}
						mSDL->displayText(dwFontImg, szText, 8, 426);
						mSDL->clearEventQueue();
					}
					break;
				default:
					break;
				}
			}
			break;
		default:
			break;
		}
	}
	mSDL->stopMusic();
	mSDL->blackScreen();
	mSDL->drawFrame();
	mSDL->freeMusic();
	mSDL->mImgArray.freeIA();
	return true;
}

void CCytadelaMain::animateMapCursor(uint32_t dwCursorImg, uint32_t *pLastTime, uint32_t *pCursorFrame, uint32_t dwNum)
{
	uint32_t dwCurTime = mSDL->getTicks();
	uint32_t dwMiliseconds = dwCurTime - *pLastTime;
	//now last time is current time minus difference between now and
	//the time when cursor should have been changed (and wasn't, cause
	//something else was beeing done in program)
	*pLastTime = dwCurTime - (dwMiliseconds % 27);
	//how many times cursor should have changed since last time
	*pCursorFrame += dwMiliseconds / 27;
	while(*pCursorFrame >= 24)
		*pCursorFrame -= 24;
	SDL_Rect sSrcRect = {26 * *pCursorFrame, 0, 26, 38};
	uint16_t dwCurPositions[7][2] = {{330, 164}, {106, 230}, {42, 108}, {170, 74}, {426, 48}, {554, 160}, {552, 282}};
	mSDL->displayImage(dwCursorImg, &sSrcRect, dwCurPositions[dwNum][0], dwCurPositions[dwNum][1]);
}

bool CCytadelaMain::gameOver()
{
	uint32_t dwBkgndImg, dwGameImg, dwOverImg;
	//load startup screen
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "gameover.tga", &dwBkgndImg))
		return false;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "game.tga", &dwGameImg))
		return false;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "over.tga", &dwOverImg))
		return false;
	if(not mSDL->loadMusic("music/GameOver.mod"))
		return false;
	uint32_t dwDiff = 0;
	bool bSecondScreen = false;
	mSDL->displayImage(dwBkgndImg, 0, 0, 0);
	mSDL->playMusic(1);
	uint32_t dwStartTime = mSDL->getTicks();
	mSDL->drawFrame();
	mSDL->clearEventQueue();
	while(dwDiff < 86000)
	{
		if(not bSecondScreen)
		{
			if(SDLK_RETURN == mSDL->readInputMenu() or dwDiff >= 40000)
			{
				uint32_t dwT0 = mSDL->getTicks(), dwY = 0;
				while(dwY < 196)
				{
					SDL_Rect sSrcRect = {0, 196 - dwY, 624, 4 + dwY};
					mSDL->displayImage(dwGameImg, &sSrcRect, 8, 43);
					sSrcRect.y = 0;
					sSrcRect.h = dwY;
					mSDL->displayImage(dwOverImg, &sSrcRect, 8, 439 - dwY);
					mSDL->drawFrame();
					dwY = ((mSDL->getTicks() - dwT0) * 196) / 1000;
				}
				mSDL->displayImage(dwGameImg, 0, 8, 43);
				mSDL->displayImage(dwOverImg, 0, 8, 243);
				mSDL->drawFrame();
				bSecondScreen = true;
				mSDL->clearEventQueue();
			}
		}
		else if(SDLK_RETURN == mSDL->readInputMenu())
			break;
		dwDiff = mSDL->getTicks() - dwStartTime;
	}
	mSDL->stopMusic();
	mSDL->freeMusic();
	mSDL->blackScreen();
	mSDL->drawFrame();
	mSDL->mImgArray.freeIA();
	return true;
}

//max 40 chars
bool CCytadelaMain::errorScreen(const char *szErrText)
{
	uint32_t uiFont;
	if(not mSDL->mImgArray.loadImg("tex/Menu.far", "configfont.tga", &uiFont))
		return false;

	mSDL->blackScreen();
	mSDL->displayText(uiFont, szErrText, 320 - (strlen(szErrText) * 8), 200);
	mSDL->displayText(uiFont, "FOR MORE DESCRIPTIVE INFORMATION ON\n"
								"  THE ERROR PLEASE SEE THE GAME'S\n"
								"          STANDARD OUTPUT", 40, 300);
	mSDL->displayText(uiFont, "     PRESS 'ENTER' TO CONTINUE", 40, 460);
	mSDL->drawFrame();

	//wait for the user's input
	while(SDLK_RETURN != mSDL->readInputMenu());

	mSDL->blackScreen();
	mSDL->drawFrame();
	mSDL->mImgArray.freeIA();
	return true;
}

void CCytadelaMain::runMain()
{
	fprintf(stdout, "Welcome to the Citadel!\n\n");
	CCytadelaConfig clConfig(mSysDirPath);
	if(not configScreen(&clConfig))
		return;
	setLevelTexts();
	//display startup screen
	if(not startupScreen())
	{
		errorScreen("ERROR: COULDN'T DISPLAY STARTUP SCREEN!");
		fprintf(stdout, "\nAbnormal program termination!\n");
		return;
	}
	mSDL->blackScreen();
	mSDL->drawFrame();
	CCytadelaMenu clMenu(&clConfig);
	CCytadelaGame clGame;
	COGLClass clOGL;
	//menu<->gameplay switching loop
	while(true)
	{
		reset(&clConfig);

		//copy savenames
		CCytadelaSave clSave;
		if(not clSave.readSaveFile(mSysDirPath))
		{
			fprintf(stderr, "Failed to read save file - writing empty one\n");
			clSave.reset();
			if(not clSave.writeSaveFile(mSysDirPath))
			{
				fprintf(stderr, "Error: Failed to write empty save file - exiting\n");
				//display information about the error that occured
				errorScreen("ERROR WHILE READING SAVE FILE!");
				break;
			}
		}
		const char *szSaveNames[5];
		for(int32_t i = 0; i < 5; ++i)
		{
			Levels eLevel = clSave.mSave[i].eLevel;
			if(eLevel == NO_LEVEL)
				szSaveNames[i] = mLocInterface->getText(LOC_SAVE_EMPTY);
			else
				szSaveNames[i] = mSaveNames[eLevel];
		}

		//init the menu
		if(not clMenu.initMenu(szSaveNames))
		{
			clMenu.freeMenu();
			fprintf(stderr, "Error: Failed to initialize menu - exiting\n");
			//display information about the error that occured
			errorScreen("ERROR WHILE INITIALIZING MENU!");
			break;
		}
		//run the menu
		MenuAction eResult = clMenu.runMenu();
		clMenu.freeMenu();
		switch(eResult)
		{
			case MENUACT_EXIT:
				fprintf(stdout, "Program terminated on user request\n");
				return;
			case MENUACT_START_GAME:
				mLevel = LEVEL01A;
				break;
			case MENUACT_LOAD_1:
			case MENUACT_LOAD_2:
			case MENUACT_LOAD_3:
			case MENUACT_LOAD_4:
			case MENUACT_LOAD_5:
			{
				uint32_t dwNum = eResult - MENUACT_LOAD_1;
				for(int32_t i = 0; i < 7; ++i)
					mCplxStatus[i] = clSave.mSave[dwNum].eCplxStatus[i];
				for(int32_t i = 0; i < 6; ++i)
					mState.wAmmo[i] = clSave.mSave[dwNum].wAmmo[i];
				mLevel = clSave.mSave[dwNum].eLevel;
				mState.bDifficult = clSave.mSave[dwNum].bDifficult;
				mActualComplex = clSave.mSave[dwNum].btActualComplex;
				mState.btBomb = clSave.mSave[dwNum].btBomb;
				mCplxCounter = clSave.mSave[dwNum].wCplxCounter;
				mState.wEnergy = clSave.mSave[dwNum].wEnergy;
				mState.wWeapons = clSave.mSave[dwNum].wWeapons;
				mState.bVisionNoise = clSave.mSave[dwNum].bVisionNoise;
				goto loaded;
			}
			break;
			case MENUACT_START_TR1:
				mLevel = TRAINING1;
				break;
			case MENUACT_START_TR2:
				mLevel = TRAINING2;
				break;
			case MENUACT_START_TR3:
				mLevel = TRAINING3;
				break;
			case MENUACT_START_TR4:
				mLevel = TRAINING4;
				break;
			case MENUACT_START_TR5:
				mLevel = TRAINING5;
				break;
			default:
				break;
		}

		gameplay: {
			//check the configuration that might have changed in the menu
			const int32_t *pScreenSize = clConfig.getScreenSize();
			mState.bDifficult = clConfig.isDifficult();
			if(not mSDL->switchVideoMode(pScreenSize[0], pScreenSize[1], MODE_3D))
			{
				if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_2D))
					break;
				//display information about the error that occured
				if(not errorScreen("ERROR WHILE SETTING VIDEO MODE!"))
					break;
				continue;
			}
			if(not clOGL.initOGL(pScreenSize[0], pScreenSize[1]))
			{
				clOGL.freeOGL();
				if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_2D))
					break;
				//display information about the error that occured
				if(not errorScreen("ERROR WHILE INITIALIZING OPENGL!"))
					break;
				continue;
			}
			if(not clGame.initGame(&clOGL, &mState, mTexts, &clConfig) or
				not clGame.runGame(mLevels[mLevel], &mState))
			{
				clGame.freeGame();
				clOGL.freeOGL();
				if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_2D))
					break;
				//display information about the error that occured
				if(not errorScreen(clGame.getLastErrorDesc()))
					break;
				continue;
			}
			clGame.freeGame();
			clOGL.freeOGL();
			if(not mSDL->switchVideoMode(CSDL_DEFAULT_SCREEN_W, CSDL_DEFAULT_SCREEN_H, MODE_2D))
				break;

			mSDL->blackScreen();
			mSDL->drawFrame();

			if(mState.wEnergy == 0)
			{
				if(not gameOver())
					errorScreen("COULDN'T DISPLAY 'GAME OVER' SCREEN :(");
				continue;
			}
			if(mState.bExitGame or mLevel >= TRAINING1)
				continue;
			if(not endLevel())
			{
				errorScreen("ERROR: COULDN'T DISPLAY LEVEL SUMMARY!");
				break;
			}
		}

		loaded:
		//player may exit the game during endLevel()
		if(mState.bExitGame or mLevel == LEVEL08A)
			continue;
		if(((mLevel - 1) % 3) == 0)
		{
			if(not chooseComplex())
			{
				errorScreen("ERROR: COULDN'T DISPLAY COMPLEX MENU!");
				break;
			}
		}
		else
			mLevel = (Levels)(mLevel + 1);
		//player may exit the game during chooseComplex()
		if(not mState.bExitGame)
			goto gameplay;
	}
	fprintf(stdout, "\nAbnormal program termination!\n");
}
